#include <random>
#include <iostream>
#include <algorithm>
#include <cassert>
#include <cmath>
#include <ANN/ANN.h>
#include "hair.h"

typedef unsigned int uint;

// ANNpoint is a pointer to ANNcoord.
inline void convertPoint(const Point3 &p, ANNpoint &_p) {
	_p[0] = p.x; _p[1] = p.y; _p[2] = p.z;
}

/// Van der Corput radical inverse in base 2 with double precision
inline double radicalInverse2Double(uint64_t n, uint64_t scramble) {
	/* Efficiently reverse the bits in 'n' using binary operations */
#if (defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 2))) || defined(__clang__)
	n = __builtin_bswap64(n);
#else
	n = (n << 32) | (n >> 32);
	n = ((n & 0x0000ffff0000ffffULL) << 16) | ((n & 0xffff0000ffff0000ULL) >> 16);
	n = ((n & 0x00ff00ff00ff00ffULL) << 8) | ((n & 0xff00ff00ff00ff00ULL) >> 8);
#endif
	n = ((n & 0x0f0f0f0f0f0f0f0fULL) << 4) | ((n & 0xf0f0f0f0f0f0f0f0ULL) >> 4);
	n = ((n & 0x3333333333333333ULL) << 2) | ((n & 0xccccccccccccccccULL) >> 2);
	n = ((n & 0x5555555555555555ULL) << 1) | ((n & 0xaaaaaaaaaaaaaaaaULL) >> 1);

	// Account for the available precision and scramble
	n = (n >> (64 - 53)) ^ (scramble & ~- (1LL << 53));

	return (double) n / (double) (1ULL << 53);
}

struct ClumpParams {
	uint clumpNum;
	Float clumpNumRatio;
	uint clumpNeighbor;
	uint clumpOverlap;
	Float clumpRadius;
	int randomSeed;
};
constexpr Float Epsilon = 1e-4f;

bool clumpHair(const ClumpParams &params, const Hair &originalHair, Hair &hair, std::vector<Clump>& clumps) {
	std::default_random_engine e;
	std::uniform_int_distribution<uint64_t> rand(params.randomSeed);

	auto getLdSamples = [&e, &rand](size_t sampleCount, int capacity) -> std::vector<size_t> {
		uint64_t scramble = rand(e);
		std::vector<size_t> samples(sampleCount);
		for(size_t i = 0; i < sampleCount; ++i)
			samples[i] = (int) std::floor(radicalInverse2Double(i, scramble) * capacity);
		return samples;
	};

	uint nHair = originalHair.hairCount, nClump;
	if(params.clumpNum > 0)
		nClump = std::min(nHair, params.clumpNum);
	else
		nClump = (uint) std::floor(nHair * params.clumpNumRatio);
	if(nClump == 0) 
		return false;
	clumps.clear();
	clumps.reserve(nClump);

	// hair geometry
	auto &vertices = originalHair.vertices;
	auto &vertexStartsFiber = originalHair.vertexStartsFiber;
	auto &verticesTexCoord = originalHair.verticesTexCoord;
	// hair index -> hair root vertice index.
	std::vector<size_t> hairRootVertIdx(nHair, -1);
	// hair vertice index -> hair index.
	std::vector<size_t> hairIndex(vertices.size());

	// ANN variables
	uint k = params.clumpNeighbor + 1; // first nearest point is root itself.
	ANNidx *nnIdx = new ANNidx[k];
	ANNdist *dists = new ANNdist[k];
	ANNpointArray hairRoots = annAllocPts(nHair, 3);
	ANNkd_tree *kdTree;

	size_t iHair = 0;
	for(int i = 0; i < vertices.size(); i++) {
		hairIndex[i] = iHair;
		/* setup root vertices for knn search. */
		if(!vertexStartsFiber[i]) {
			if(i + 1 < vertices.size() && vertexStartsFiber[i + 1])
				++iHair;
			continue;
		}
		hairRootVertIdx[iHair] = i;
		convertPoint(vertices[i], hairRoots[iHair]);
	}

	kdTree = new ANNkd_tree(hairRoots, nHair, 3);

	std::vector<bool> isHairClump(nHair, false);
	std::vector<size_t> idxSamples = getLdSamples(nClump, nHair);
	for(uint i = 0; i < nClump; i++)
		isHairClump[idxSamples[i]] = true;

	int clumpIdx = 0;
	/* knn search and calculate clump radius and uv bound. */
	for(auto idxIter = idxSamples.begin(); idxIter != idxSamples.end(); idxIter++)
	{
		kdTree->annkSearch(
			hairRoots[*idxIter],
			k,
			nnIdx,
			dists,
			Epsilon
		);
		Point2 texBoundU = { 1.0f, 0.0f }, texBoundV = { 1.0f, 0.0f };
		uint i = 1, nOverlap = 0;
		for(; i < k; i++) {
			if(dists[i] > params.clumpRadius) break;
			/* expands uv bound of each clump, until encounters given numbers of other clumps.*/
			if(isHairClump[nnIdx[i]] && ++nOverlap > params.clumpOverlap) break;
			auto texCoord = verticesTexCoord[hairRootVertIdx[nnIdx[i]]];
			texBoundU.x = std::min(texBoundU.x, texCoord.x);
			texBoundU.y = std::max(texBoundU.y, texCoord.x);
			texBoundV.x = std::min(texBoundV.x, texCoord.y);
			texBoundV.y = std::max(texBoundV.y, texCoord.y);
		}
		/*  if there is only a single hair in the clump, we simply discard it for lack of hair.*/
		if(i == 1) {
			isHairClump[*idxIter] = false;
			continue;
		}
		clumps.push_back({(Float)std::sqrt(dists[i - 1]), texBoundU, texBoundV });
	}
	delete kdTree;
	delete[] nnIdx;
	delete[] dists;
	annDeallocPts(hairRoots);

	std::vector<Point3> vertices_;
	std::vector<bool> vertexStartsFiber_;
	std::vector<Point2> verticesTexCoord_;
	uint actualNumClump = 0;
	/* export vertices of each clump back. */
	for(size_t i = 0; i < nClump; i++) {
		if(!isHairClump[idxSamples[i]]) continue;
		actualNumClump++;
		size_t j = hairRootVertIdx[idxSamples[i]];
		do {
			vertices_.push_back(vertices[j]);
			vertexStartsFiber_.push_back(vertexStartsFiber[j]);
			verticesTexCoord_.push_back(verticesTexCoord[j]);
			j++;
		} while(j < vertices.size() && !vertexStartsFiber[j]);
	}

	hair = Hair { 
		std::move(vertices_),
		std::move(vertexStartsFiber_),
		std::move(verticesTexCoord_),
		actualNumClump 
	};
	return true;
}

int main(int argc, char* argv[])
{
	if (argc < 5) {
		std::cout << "Args: \n"
				  << " #1 input path\n"
				  << " #2 output path\n"
				  << " #3 hair num / ratio after clump\n"
				  << " #4 queried neighbor\n"
				  << " #5 (optional) max overlap clump (1)\n"
				  << " #6 (optional) max query radius (inf)\n"
				  << " #7 (optional) random seed (0)\n";
		return -1;
	}
	std::string inputPath(argv[1]);
	std::string outputPath(argv[2]);

	ClumpParams params;

	std::string _clumpNumOrRatio(argv[3]);
	if (_clumpNumOrRatio.find('.')!=-1) {
		params.clumpNum = 0;
		params.clumpNumRatio = std::stof(_clumpNumOrRatio);
	}
	else {
		params.clumpNum = std::stoi(_clumpNumOrRatio);
		params.clumpNumRatio = 0.f;
	}
	params.clumpNeighbor = std::stoi(argv[4]);
	params.clumpOverlap = argc > 5 ? std::stoi(argv[5]) : 1;
	params.clumpRadius = argc > 6 ? std::stof(argv[6]) : 1e10f;
	params.randomSeed = argc > 7 ? std::stoi(argv[7]) : 0;

	Hair hair;
	if(!importHair(inputPath, hair)) {
		std::cerr << "Failed to import hair.\n";
		return -1;
	}

	Hair clumpedHair;
	std::vector<Clump> clumps;
	if(!clumpHair(params, hair, clumpedHair, clumps)) {
		std::cerr << "Failed to clump hair.\n";
		return -1;
	}
	if(!exportHairBinary(outputPath, clumpedHair)) {
		std::cerr << "Failed to export hair.\n";
		return -1;
	}
	if(!exportInfo(outputPath + ".info", clumps)) {
		std::cerr << "Failed to export clump info.\n";
		return -1;
	}
	return 0;
}